@@ -1,5 +1,3 @@
-#!/usr/bin/env python3
-
 import argparse
 
 flags = argparse.ArgumentParser(description='Take control of your Palette.')
@@ -13,6 +11,9 @@ flags.add_argument('--debug', default=False, action='store_true',
 flags.add_argument('--verbose', default=False, action='store_true',
                    help='include more detailed status output')
 
+flags.add_argument('--layouts', default=[], type=str, nargs='*',
+                   help='launch with the specified layouts')
+
 flags.add_argument('config_paths', type=str, metavar='CONFIG', nargs='*',
                    help='configuration files to load')
 
